//
// Copyright (c) 2019 INRIA
//

#ifndef __pinocchio_math_casadi_hpp__
#define __pinocchio_math_casadi_hpp__

#include "pinocchio/macros.hpp"

PINOCCHIO_PRAGMA_DEPRECATED_HEADER(pinocchio/math/casadi.hpp,pinocchio/autodiff/casadi.hpp)

#include "pinocchio/autodiff/casadi.hpp"

#endif // #ifndef __pinocchio_math_casadi_hpp__
